# Builtin modules
from __future__ import annotations
from time import monotonic
from typing import Any, Union, List, Dict, Optional
# Third party modules
# Local modules
from .abcs import T_Request, T_Client
# Program
class Request(T_Request):
	def __init__(self, client:T_Client, id:Any, method:str, args:List[Any], kwargs:Dict[str, Any], path:str="/",
	convertNumbers:Optional[str]=None) -> None:
		self._client         = client
		self._id             = id
		self._method         = method
		self._args           = args
		self._kwargs         = kwargs
		self._path           = path
		self._convertNumbers = convertNumbers
		#
		self._requestTime  = monotonic()
		self._responseTime = 0.0
		self._uid          = ""
		self._done         = False
		self._success      = False
		self._response     = None
	def _get(self) -> None:
		self._client._get(self._id)
		return None
	def _parseResponse(self, id:Union[int, str], isSuccess:bool, result:Any, uid:str) -> None:
		self._done = True
		self._responseTime = monotonic()
		self._uid = uid
		self._success = isSuccess
		self._response = result
	def _dumps(self) -> Any:
		if self._method:
			r:Any
			if self._client.requestProtocol == "JSONRPC-2":
				r = {
					"jsonrpc":"2.0",
					"params":self._kwargs or self._args,
					"method":self._method,
					"id":self._id,
				}
				if self._convertNumbers is not None:
					r["convertNumbers"] = self._convertNumbers
			elif self._client.requestProtocol == "JSONRPC-P":
				r = {
					"jsonrpc":"python",
					"args":self._args,
					"kwargs":self._kwargs,
					"method":self._method,
					"id":self._id,
				}
				if self._convertNumbers is not None:
					r["convertNumbers"] = self._convertNumbers
			elif self._client.requestProtocol == "FSP":
				r = (self._id, self._method, self._args, self._kwargs)
			elif self._client.requestProtocol == "OldFSProtocol":
				r = (self._id, self._method, self._args, self._kwargs)
			else:
				raise RuntimeError
			return r
		return None
	def get(self) -> Any:
		if not self._done:
			self._get()
		return self._response
	def getDelay(self) -> float:
		if not self._done:
			self._get()
		return self._responseTime - self._requestTime
	def getID(self) -> Any:
		return self._id
	def isDone(self) -> bool:
		return self._done
	def getUID(self) -> str:
		if not self._done:
			self._get()
		return self._uid
	def isSuccess(self) -> bool:
		if not self._done:
			self._get()
		return self._success



# Builtin modules
import re, ssl, traceback, socket, errno, codecs
from time import monotonic
from typing import Tuple, Dict, Union, Callable, NoReturn, Optional, cast
from selectors import DefaultSelector, EVENT_READ, EVENT_WRITE
from math import ceil
# Third party modules
import fsPacker
# Local modules
from . import __version__
from .utils import Headers, deflate
from .exceptions import SocketError, MessageError
from .abcs import (T_Socket, T_Client, T_BaseClientSocket, T_HTTPClientSocket, T_StringClientSocket, T_FSPackerClientSocket,
T_OldFSPackerClientSocket, SSLContext, T_SocketBindAddress)
# Program
NOT_CONNECTED = 0
CONNECTING    = 1
CONNECTED     = 2

class BaseClientSocket(T_BaseClientSocket):
	def __init__(self, client:T_Client, protocol:str, target:Union[str, Tuple[str, int], Tuple[str, int, int, int]],
	bind:Optional[T_SocketBindAddress], connectTimeout:Union[int, float], transferTimeout:Union[int, float], ssl:bool,
	sslHostname:Optional[str]) -> None:
		assert protocol in ["TCPv4", "TCPv6", "IPC"], "Unsupported protocol"
		self.client            = client
		self.protocol          = protocol
		self.target            = target
		self.connectTimeout    = float(connectTimeout)
		self.transferTimeout   = float(transferTimeout)
		self.ssl               = ssl
		self.sslHostname       = sslHostname
		self.bind              = bind
		self.log               = self.client.log.getChild("socket")
		self.signal            = self.client.signal
		self.poll              = DefaultSelector()
		self.sockFD            = None
		self._reset()
		self.log.debug("Initialized")
	def _connect(self, initial:bool=False) -> bool:
		if not hasattr(self, "sock"): return True
		if initial:
			if self.protocol == "TCPv4":
				self.log.info("Connecting to {}:{} ..".format(*self.target))
			elif self.protocol == "TCPv6":
				self.log.info("Connecting to {} {} [{}/{}] ..".format(*self.target))
			elif self.protocol == "IPC":
				self.log.info("Connecting to {} ..".format(self.target))
		cerr = self.sock.connect_ex(self.target)
		if cerr in [errno.EAGAIN, errno.EINPROGRESS]: # errno.ENETUNREACH, errno.EADDRNOTAVAIL
			return False
		elif cerr in [0, errno.EISCONN]:
			if self.ssl:
				self.sslTimer = monotonic()
				sslCtx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
				sslCtx.set_ciphers(
					":".join([
						"ECDHE-ECDSA-AES256-GCM-SHA384",
						"ECDHE-RSA-AES256-GCM-SHA384",
						"DHE-RSA-AES256-GCM-SHA384",
						"ECDHE-ECDSA-CHACHA20-POLY1305",
						"ECDHE-RSA-CHACHA20-POLY1305",
						"DHE-RSA-CHACHA20-POLY1305",
						"ECDHE-ECDSA-AES256-SHA384",
						"ECDHE-RSA-AES256-SHA384",
						"DHE-RSA-AES256-SHA256",
						"ECDHE-ECDSA-AES256-SHA",
						"ECDHE-RSA-AES256-SHA",
						"DHE-RSA-AES256-SHA",
						"RSA-PSK-AES256-GCM-SHA384",
						"DHE-PSK-AES256-GCM-SHA384",
						"RSA-PSK-CHACHA20-POLY1305",
						"DHE-PSK-CHACHA20-POLY1305",
						"ECDHE-PSK-CHACHA20-POLY1305",
						"AES256-GCM-SHA384",
						"PSK-AES256-GCM-SHA384",
						"PSK-CHACHA20-POLY1305",
						"ECDHE-PSK-AES256-CBC-SHA384",
						"ECDHE-PSK-AES256-CBC-SHA",
						"SRP-RSA-AES-256-CBC-SHA",
						"SRP-AES-256-CBC-SHA",
						"RSA-PSK-AES256-CBC-SHA384",
						"DHE-PSK-AES256-CBC-SHA384",
						"RSA-PSK-AES256-CBC-SHA",
						"DHE-PSK-AES256-CBC-SHA",
						"AES256-SHA",
						"PSK-AES256-CBC-SHA384",
						"PSK-AES256-CBC-SHA"
					])
				)
				self.sock = cast(SSLContext, sslCtx).wrap_socket(
					self.sock, False, False, server_hostname=self.sslHostname or self.target[0]
				)
				self.log.debug("Doing SSL handshake..")
			self.connectionStatus = CONNECTING
			self._setMask(EVENT_READ | EVENT_WRITE)
			if not self.ssl:
				self.connectionStatus = CONNECTED
				connectionDelay =  monotonic()-self.timeoutTimer
				self.timeoutTimer = monotonic()
				self.log.info("Connected in {:.3F} sec".format(connectionDelay))
		elif not initial:
			self._raiseSocketError("Connection failed {}[{}]".format(errno.errorcode[cerr], cerr))
		return False
	def _createSocket(self) -> None:
		if hasattr(self, "sock"): return None
		self._reset()
		self.sock = cast(T_Socket, socket.socket(
			{
				"TCPv4":socket.AF_INET,
				"TCPv6":socket.AF_INET6,
				"IPC":socket.AF_UNIX,
			}[self.protocol]
		))
		self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		self.sock.setblocking(False)
		if self.bind is not None:
			try:
				self.sock.bind(self.bind)
			except OSError as err:
				if err.errno == errno.EADDRINUSE:
					self.log.warn("Bind address {} already in use, switch to automatic", self.bind)
					# elif err.errno == errno.EADDRNOTAVAIL:
					# 	self.log.warn("Bind address {} not available, switch to automatic", self.bind)
				else:
					self._raiseSocketError("Error during binding socket to {}: {}".format(self.bind, err))
				self.bind = None
				self._reset()
				return self._createSocket()
		newSockFD = self.sock.fileno()
		if self.sockFD and self.sockFD != newSockFD:
			self.poll.unregister(self.sockFD)
			self.sockFD = None
		if self.sockFD and self.sockFD == newSockFD:
			self.poll.modify(newSockFD, EVENT_READ)
		else:
			self.poll.register(newSockFD, EVENT_READ)
		self.sockFD = newSockFD
		self.mask = EVENT_READ
		self.log.info("Socket created [FD:{}]".format(self.sockFD))
	def _doSSLHandshake(self) -> bool:
		try:
			self.sock.do_handshake()
		except ssl.SSLWantReadError:
			self._setMask(EVENT_READ)
			return False
		except ssl.SSLWantWriteError:
			self._setMask(EVENT_WRITE)
			return False
		except Exception:
			self._raiseSocketError("SSL Handsake error")
		self._setMask(EVENT_READ | EVENT_WRITE)
		self.log.info("Connected in {:.3F} sec [SSL: {:.3F} sec]".format(
			monotonic()-self.timeoutTimer,
			monotonic()-self.sslTimer,
		))
		self.connectionStatus = CONNECTED
		return False
	def _haveRead(self) -> bool:
		if not hasattr(self, "sock"): return True
		data = b""
		if self.ssl and self.connectionStatus == CONNECTING:
			return self._doSSLHandshake()
		try:
			data = self.sock.recv(16<<20)
		except ssl.SSLWantReadError:
			self._setMask(EVENT_READ)
			return False
		except ssl.SSLWantWriteError:
			self._setMask(EVENT_WRITE)
			return False
		except BlockingIOError:
			pass
		except ConnectionRefusedError:
			self._raiseSocketError("Connection {}".format("broken" if self.connectionStatus == CONNECTED else "refused"))
		except:
			self._raiseSocketError("Unknown error: {}".format(traceback.format_exc()))
		if not data:
			self._raiseSocketError("Connection broken")
		self.readBuffer += data
		if self.log.isFiltered("TRACE"):
			self.log.debug("Read {} bytes [{} bytes in buffer]:\n{!r}".format(
				len(data),
				len(self.readBuffer),
				data,
			))
		elif self.log.isFiltered("DEBUG"):
			self.log.debug("Read {} bytes [{} bytes in buffer]".format(
				len(data),
				len(self.readBuffer),
			))
		if self.parseReadBuffer():
			return True
		self.timeoutTimer = monotonic()
		return False
	def _haveWrite(self) -> bool:
		if not hasattr(self, "sock"): return True
		sentLength = 0
		if self.connectionStatus == NOT_CONNECTED:
			return self._connect()
		if self.connectionStatus == CONNECTING and self.ssl:
			return self._doSSLHandshake()
		if not self.writeBuffer:
			self._setMask(EVENT_READ)
			return False
		try:
			sentLength = self.sock.send(self.writeBuffer)
		except ssl.SSLWantReadError:
			self._setMask(EVENT_READ)
		except ssl.SSLWantWriteError:
			self._setMask(EVENT_WRITE)
		except BrokenPipeError:
			self._raiseSocketError("Connection broken")
		if sentLength:
			self.timeoutTimer = monotonic()
			if self.log.isFiltered("TRACE"):
				self.log.trace("Sent {} bytes [{} still in buffer]:\n{!r}".format(
					sentLength,
					len(self.writeBuffer)-sentLength,
					self.writeBuffer[:sentLength],
				))
			elif self.log.isFiltered("DEBUG"):
				self.log.debug("Sent {} bytes [{} still in buffer]".format(
					sentLength,
					len(self.writeBuffer)-sentLength,
				))
			self.writeBuffer = self.writeBuffer[sentLength:]
		if not self.writeBuffer:
			self._setMask(EVENT_READ)
		return False
	def _raiseSocketError(self, err:str) -> NoReturn:
		self._reset()
		self.log.error(err)
		raise SocketError(err) from None
	def _raiseMessageError(self, err:str) -> NoReturn:
		self.log.error(err)
		raise MessageError(err) from None
	def _reset(self) -> None:
		self.readBuffer = b""
		self.writeBuffer = b""
		self.connectionStatus = NOT_CONNECTED
		self.timeoutTimer = 0.0
		if hasattr(self, "sock"):
			try: self.sock.shutdown(socket.SHUT_RDWR)
			except: pass
			try: self.sock.close()
			except: pass
			del self.sock
		self.mask = EVENT_READ
		self.sslTimer = 0.0
	def _setMask(self, newMask:int) -> None:
		if self.sockFD and self.mask != newMask:
			self.mask = newMask
			self.log.trace("New mask: {}", newMask)
			self.poll.modify(self.sockFD, newMask)
		return None
	def _write(self, data:bytes) -> None:
		self.writeBuffer += data
		if self.connectionStatus == CONNECTED:
			self._setMask(EVENT_READ | EVENT_WRITE)
	def close(self) -> None:
		if hasattr(self, "sock"):
			self.log.trace("Closing [FD:{}]", self.sockFD)
			self.mask = EVENT_READ
			self._reset()
			self.log.debug("Closed")
	def connect(self) -> None:
		if self.connectionStatus != NOT_CONNECTED:
			return
		self._createSocket()
		self.timeoutTimer = monotonic()
		self._setMask(EVENT_WRITE)
		self._connect(initial=True)
		if self.connectionStatus != CONNECTED:
			self.loop(self._isConnected)
	def _isConnected(self) -> bool:
		return self.connectionStatus == CONNECTED
	def isAlive(self) -> bool:
		return not (
			(self.connectionStatus == NOT_CONNECTED) or
			(self.connectionStatus == CONNECTING and monotonic()-self.timeoutTimer > self.connectTimeout) or
			(self.connectionStatus == CONNECTED and monotonic()-self.timeoutTimer > self.transferTimeout)
		)
	def loop(self, whileFn:Callable[[], bool]) -> None:
		checkTimer = monotonic()
		try:
			while whileFn():
				self.signal.check()
				if monotonic()-checkTimer >= 1:
					checkTimer = monotonic()
					if not self.isAlive():
						self._raiseSocketError("Timeout")
				pollReturn = self.poll.select(1)
				for sk, pollBitmask in pollReturn:
					if sk.fd == self.sockFD:
						if pollBitmask & EVENT_READ and self._haveRead():
							return None
						if pollBitmask & EVENT_WRITE and self._haveWrite():
							return None
		except:
			self.close()
			raise
	def parseReadBuffer(self) -> bool:
		raise RuntimeError

class HTTPClientSocket(BaseClientSocket, T_HTTPClientSocket):
	defaultHeaders = {
		"User-Agent":"Fusion Solutions RPC Client v{}".format(__version__),
		"Accept":"*/*",
		"Connection":"Keep-Alive",
	}
	def __init__(self, client:T_Client, protocol:str, target:Union[str, Tuple[str, int], Tuple[str, int, int, int]],
	bind:Optional[T_SocketBindAddress], connectTimeout:Union[int, float], transferTimeout:Union[int, float], ssl:bool=False,
	sslHostname:Optional[str]=None, extraHeaders:Dict[str, str]={}, disableCompression:bool=False) -> None:
		super().__init__(client, protocol, target, bind, connectTimeout, transferTimeout, ssl, sslHostname)
		self.headers = Headers(self.defaultHeaders)
		self.headers.update(extraHeaders)
		if not disableCompression:
			self.headers.update({"Accept-Encoding":"deflate"})
		return None
	def parseReadBuffer(self) -> bool:
		endLine = b"\r\n"
		pos = self.readBuffer.find(endLine*2)
		if pos == -1:
			endLine = b"\n"
			pos = self.readBuffer.find(endLine*2)
		elif pos > 4096:
			self._raiseMessageError("HTTP headers are too long")
		endLineLen = len(endLine)
		while pos != -1:
			rawData = self.readBuffer[pos+endLineLen*2:]
			rawHeaders = self.readBuffer[:pos].decode("ISO-8859-1").split(endLine.decode("ISO-8859-1"))
			if not rawHeaders:
				self._raiseMessageError("Invalid HTTP headers")
			httpResponse = rawHeaders.pop(0).split(" ")
			if len(httpResponse) < 2:
				self._raiseMessageError("Invalid HTTP response code")
			# if httpResponse[1] == "503":
			# 	self._raiseSocketError("Server offline")
			# elif httpResponse[1] != "200":
			# 	self._raiseSocketError("Request failure")
			headers = Headers()
			for rawHeader in rawHeaders:
				s = rawHeader.find(":")
				if s <= 0 or s > 64:
					self._raiseMessageError("Header key too long")
					return
				headers[ rawHeader[:s].strip() ] = rawHeader[s+1:].strip()
			cLengthStr = headers.get("content-length", "")
			cLength:Optional[int] = None
			if cLengthStr:
				if not cLengthStr.isdigit():
					self._raiseMessageError("Invalid HTTP header value for content-length")
				cLength = int(cLengthStr)
				if cLength < 0:
					self._raiseMessageError("Invalid HTTP header value for content-length")
			elif "chunked" not in headers.get("transfer-encoding", "").lower():
				self._raiseMessageError("Invalid HTTP transfer encoding. Not chunked and no content-length given.")
			cEncoding = headers.get("content-encoding", "")
			if cEncoding == "":
				compression = False
			elif cEncoding == "deflate":
				compression = True
			else:
				self._raiseMessageError("Request HTTP encoding not supported")
			cType = headers.get("content-type", "").split(";")[0].strip().lower()
			if cType == "":
				self._raiseMessageError("Invalid HTTP header value for content-type")
			cTypeS = re.findall("charset=([a-z0-9-]*)", headers.get("content-type", ""), re.I)
			charset = cTypeS[0] if len(cTypeS) == 1 else "iso-8859-1"
			try:
				codecs.lookup(charset)
			except LookupError:
				self._raiseMessageError("Not supported charset")
			headers["content-type"] = cType
			#
			payload = b""
			if cLength is None:
				# Chunked: data_size:int + endline:bytes + data:bytes
				rawDataCache = rawData
				EOC = False
				pos = rawDataCache.find(endLine)
				cLength = 0
				while pos != -1:
					cLength += endLineLen
					if pos == 0:
						# end of chunks
						EOC = True
						break
					rawChunk = rawDataCache[:pos]
					rawDataCache = rawDataCache[pos+endLineLen:]
					try:
						chunkLength = int(rawChunk, 16)
					except:
						self._raiseMessageError("Invalid HTTP chunk length")
					if chunkLength > 0xFFFFFF:
						self._raiseMessageError("HTTP chunk too big")
					if chunkLength > len(rawDataCache):
						return False
					payload += rawDataCache[:chunkLength]
					cLength += chunkLength
					rawDataCache = rawDataCache[chunkLength:]
					pos = rawDataCache.find(endLine)
				if not EOC:
					return False
			else:
				if len(rawData) < cLength:
					return False
				payload = rawData[:cLength]
			assert isinstance(cLength, int)
			self.readBuffer = self.readBuffer[pos+endLineLen*2+cLength:]
			#
			if compression:
				try:
					payload = deflate.decompress(payload)
				except:
					self._raiseMessageError("Invalid response content")
			if self.log.isFiltered("TRACE"):
				self.log.trace("Payload [len: {}]: {}", len(payload), payload)
			self.client._parseResponse(payload, headers, charset)
			pos = self.readBuffer.find(endLine*2)
		return False
	def send(self, payload:bytes=b"", path:str="/", headers:Dict[str, str]={}) -> None:
		rawHeader = "{} {} HTTP/1.1\r\n".format("POST" if payload else "GET", path)
		if payload:
			headers["content-length"] = str(len(payload))
		rawHeader += self.headers.dumps(extend=headers)
		rawHeader += "\r\n\r\n"
		self._write( rawHeader.encode("iso-8859-1") + payload )

class StringClientSocket(BaseClientSocket, T_StringClientSocket):
	def parseReadBuffer(self) -> bool:
		pos = self.readBuffer.find(b"\n")
		while pos != -1:
			response = self.readBuffer[:pos]
			self.readBuffer = self.readBuffer[pos+1:]
			self.client._parseResponse(response)
			pos = self.readBuffer.find(b"\n")
		return False
	def send(self, payload:bytes) -> None:
		self._write( payload + b"\n" )
		return None

class FSPackerClientSocket(BaseClientSocket, T_FSPackerClientSocket):
	def parseReadBuffer(self) -> bool:
		while not self.signal.get():
			rbl = len(self.readBuffer)
			if rbl == 0:
				return False
			li = self.readBuffer[0]
			indicatorlength = 0
			messageLength = 0
			if li < 0xFD:
				indicatorlength = 1
				messageLength = li
			elif li == 0xFD:
				if rbl > 3:
					indicatorlength = 3
					messageLength = int.from_bytes(self.readBuffer[1:3], "little")
			elif li == 0xFE:
				if rbl > 4:
					indicatorlength = 4
					messageLength = int.from_bytes(self.readBuffer[1:4], "little")
			elif li == 0xFF:
				if rbl > 5:
					indicatorlength = 5
					messageLength = int.from_bytes(self.readBuffer[1:5], "little")
			
			if indicatorlength == 0:
				self._raiseMessageError("Got invalid response")
			if rbl < indicatorlength:
				return False
			payloadLength = indicatorlength+messageLength
			if rbl < payloadLength:
				return False
			response = self.readBuffer[indicatorlength:payloadLength]
			self.readBuffer = self.readBuffer[payloadLength:]
			self.client._parseResponse(response)
		return False
	def send(self, payload:bytes) -> None:
		responseLength = len(payload)
		if responseLength < 0xFD:
			content = responseLength.to_bytes(1, "little") + payload
		elif responseLength <= 0xFFFF:
			content = b"\xFD" + responseLength.to_bytes(2, "little") + payload
		elif responseLength <= 0xFFFFFF:
			content = b"\xFE" + responseLength.to_bytes(3, "little") + payload
		elif responseLength <= 0xFFFFFFFF:
			content = b"\xFF" + responseLength.to_bytes(4, "little") + payload
		else:
			raise fsPacker.PackingError("Too big data to pack")
		self._write(content)
		return None

class OldFSProtocolClientSocket(BaseClientSocket, T_OldFSPackerClientSocket):
	def _readResponse(self) -> Tuple[int, bytes]:
		fl = len(self.readBuffer)
		if not fl:
			return 0, b""
		li = self.readBuffer[0]
		if li == 0:
			return -1, b""
		if fl < li:
			return 0, b""
		l = int.from_bytes(self.readBuffer[ 1:1+li ], "little")
		tl = 1+li+l
		if fl < tl:
			return 0, b""
		return tl, self.readBuffer[ 1+li:tl ]
	def _encodeRequest(self, buffer:bytes) -> bytes:
		l = len(buffer)
		li = ceil(l.bit_length() / 8)
		return li.to_bytes(1, "little") + l.to_bytes(li, "little") + buffer
	def parseReadBuffer(self) -> bool:
		while not self.signal.get():
			if not self.readBuffer:
				return False
			messageLength, response = self._readResponse()
			if messageLength == -1:
				return True
			if messageLength == 0:
				return False
			self.readBuffer = self.readBuffer[messageLength:]
			self.client._parseResponse(response)
		return False
	def send(self, payload:bytes) -> None:
		self._write( self._encodeRequest( payload ) )
		return None



# Builtin modules
# Local modules
# Program
class BaseRPCError(Exception):
	def __init__(self, typ:str, message:str):
		self.message = "{}: {}".format(typ, message)
		super().__init__(self.message)

class InitializationError(BaseRPCError):
	def __init__(self, message:str):
		super().__init__("Initialization error", message)

class SocketError(BaseRPCError):
	def __init__(self, message:str):
		super().__init__("Socket error", message)

class MessageError(BaseRPCError):
	def __init__(self, message:str):
		super().__init__("Message error", message)

class RequestError(BaseRPCError):
	def __init__(self, message:str):
		super().__init__("Request error", message)

class ResponseError(BaseRPCError):
	def __init__(self, message:str):
		super().__init__("Response error", message)



__version__ = "0.2.0"
__doc__ = """
Python client for Fusion Explorer API v{}
Copyright (C) 2021 Fusion Solutions KFT <contact@fusionsolutions.io>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/lgpl.txt>.
""".format(__version__)
from .client import Client
from .requestObj import Request
from .exceptions import BaseRPCError, InitializationError, SocketError, MessageError, RequestError, ResponseError
from .abcs import T_Client, T_Request
__all__ = ("Client", "Request", "BaseRPCError", "InitializationError", "SocketError", "MessageError", "RequestError",
"ResponseError", "T_Client", "T_Request")



# Builtin modules
from __future__ import annotations
import json, weakref, pickle
from lz4.frame import decompress # type: ignore
from typing import Any, Union, List, Dict, Tuple, Optional, Callable, cast
# Third party modules
import fsPacker
from fsLogger import Logger, T_Logger
from fsSignal import HardSignal, KillSignal, T_Signal
# Local modules
from .exceptions import InitializationError, MessageError, RequestError, ResponseError, BaseRPCError
from .utils import iterSplit
from .clientSocket import HTTPClientSocket, StringClientSocket, FSPackerClientSocket, OldFSProtocolClientSocket
from .abcs import (T_Request, T_Client, T_Headers, T_HTTPClientSocket, T_BaseClientSocket_send_default, T_SocketBindAddress,
T_BaseClientSocket_send_http, T_StringClientSocket, T_FSPackerClientSocket, T_OldFSPackerClientSocket)
from .requestObj import Request
# Program
_dumpRequest:Callable[[T_Request], Any] = lambda x: x._dumps()

class Client(T_Client):
	max_bulk_request:int = 0xFF
	FSPACKER_VERSION:int = fsPacker.HIGHEST_VERSION
	def __init__(self, protocol:str, target:Union[str, Tuple[str, int], Tuple[str, int, int, int]],
	connectTimeout:Union[int, float]=15, transferTimeout:Union[int, float]=320, retryCount:int=10,
	retryDelay:Union[int, float]=5, ssl:bool=False, sslHostname:Optional[str]=None, httpHost:Optional[str]=None,
	extraHttpHeaders:Dict[str, str]={}, disableCompression:bool=False, useBulkRequest:bool=True,
	convertNumbers:Optional[str]=None, bind:Optional[T_SocketBindAddress]=None, log:Optional[T_Logger]=None,
	signal:Optional[T_Signal]=None) -> None:
		self.protocol           = protocol
		self.target             = target
		self.connectTimeout     = float(connectTimeout)
		self.transferTimeout    = float(transferTimeout)
		self.retryCount         = retryCount
		self.retryDelay         = float(retryDelay)
		self.ssl                = ssl
		self.sslHostname        = sslHostname
		self.httpHost           = httpHost
		self.extraHttpHeaders   = extraHttpHeaders
		self.disableCompression = disableCompression
		self.useBulkRequest     = useBulkRequest
		self.convertNumbers     = convertNumbers
		self.bind               = bind
		self.log                = log or Logger("RPCClient")
		self.signal             = signal or HardSignal()
		self.id                 = 0
		self.requests           = weakref.WeakValueDictionary()
		self.socketErrors       = 0
		if self.connectTimeout < 0:
			raise InitializationError("`connectTimeout` must be greater than 0")
		if self.transferTimeout < 0:
			raise InitializationError("`transferTimeout` must be greater than 0")
		if self.retryCount < 0:
			raise InitializationError("`retryCount` must be positive integer")
		if self.retryDelay < 0:
			raise InitializationError("`retryDelay` must be greater or equal than 0")
		if self.convertNumbers is not None and self.convertNumbers not in ["default", "none", "hex", "str"]:
			raise InitializationError("When `convertNumbers` is set, then most be the following one: default, none, hex or str")
		self._initializeProtocol()
	def _initializeProtocol(self) -> None:
		if ":" not in self.protocol:
			raise InitializationError("Invalid protocol")
		protocolParts = self.protocol.split(":")
		if len(protocolParts) != 3:
			raise InitializationError("Invalid protocol")
		self.socketProtocol, self.messageProtocol, self.requestProtocol = protocolParts
		if self.socketProtocol == "IPC":
			self.target = cast(str, self.target)
			if not isinstance(self.target, str):
				raise InitializationError("Target must be path")
		elif self.socketProtocol == "TCPv4":
			self.target = cast(Tuple[str, int], self.target)
			if not isinstance(self.target, (tuple, list)) or len(self.target) != 2 or \
			not isinstance(self.target[0], str) or not isinstance(self.target[1], int):
				raise InitializationError("Target must be address and port in a list\tuple")
		elif self.socketProtocol == "TCPv6":
			self.target = cast(Tuple[str, int, int, int], self.target)
			if  not isinstance(self.target, (tuple, list)) or len(self.target) != 4 or \
			not isinstance(self.target[0], str) or not isinstance(self.target[1], int) or \
			not isinstance(self.target[2], int) or not isinstance(self.target[3], int):
				raise InitializationError("Target must be address, port, flow info and scope id in a list\tuple")
		else:
			raise InitializationError("Unsupported socket protocol")
		if self.messageProtocol not in ["HTTP", "STR", "FSP", "OldFSProtocol"]:
			raise InitializationError("Unsupported message protocol")
		if self.requestProtocol not in ["JSONRPC-2", "JSONRPC-P", "FSP", "OldFSProtocol"]:
			raise InitializationError("Unsupported request protocol")
		if self.messageProtocol == "OldFSProtocol":
			if self.requestProtocol != "OldFSProtocol":
				raise InitializationError("Unsupported request protocol")
			self.useBulkRequest = False
		if self.log.isFiltered("TRACE"):
			self.log.debug(
				"Protocol initialized  [sockProt: {}][msgProt: {}][reqProt: {}][target: {}][SSL: {}]",
				self.socketProtocol, self.messageProtocol, self.requestProtocol, self.target, self.ssl
			)
		elif self.log.isFiltered("DEBUG"):
			self.log.debug("Protocol initialized")
	def __enter__(self) -> T_Client:
		return self
	def __exit__(self, type:Any, value:Any, traceback:Any) -> None:
		self.close()
	def __getstate__(self) -> Dict[str, Any]:
		return {
			"target":            self.target,
			"protocol":          self.protocol,
			"connectTimeout":    self.connectTimeout,
			"transferTimeout":   self.transferTimeout,
			"retryCount":        self.retryCount,
			"retryDelay":        self.retryDelay,
			"ssl":               self.ssl,
			"sslHostname":       self.sslHostname,
			"httpHost":          self.httpHost,
			"extraHttpHeaders":  self.extraHttpHeaders,
			"disableCompression":self.disableCompression,
			"useBulkRequest":    self.useBulkRequest,
			"convertNumbers":    self.convertNumbers,
			"bind":              self.bind,
			"log":               self.log,
			"signal":            self.signal,
		}
	def __setstate__(self, states:Dict[str, Any]) -> None:
		self.target             = states["target"]
		self.protocol           = states["protocol"]
		self.connectTimeout     = states["connectTimeout"]
		self.transferTimeout    = states["transferTimeout"]
		self.retryCount         = states["retryCount"]
		self.retryDelay         = states["retryDelay"]
		self.ssl                = states["ssl"]
		self.sslHostname        = states["sslHostname"]
		self.httpHost           = states["httpHost"]
		self.extraHttpHeaders   = states["extraHttpHeaders"]
		self.disableCompression = states["disableCompression"]
		self.useBulkRequest     = states["useBulkRequest"]
		self.convertNumbers     = states["convertNumbers"]
		self.bind               = states["bind"]
		self.log                = states["log"]
		self.signal             = states["signal"]
		self.id                 = 0
		self.requests           = weakref.WeakValueDictionary()
	def _connect(self, sendOlderRequests:bool=True) -> None:
		self.socket.connect()
		self.log.info("Connected")
		if sendOlderRequests:
			objs:List[T_Request] = list(filter(lambda x: not x.isDone(), self.requests.values()))
			if objs:
				objs.sort(key=lambda x: x._requestTime)
				self.log.info("Sending {} previous requests", len(objs))
				chunkToSend:Any
				if self.useBulkRequest and len(objs) > 1:
					for chunk in iterSplit(list(map(_dumpRequest, objs)), self.max_bulk_request):
						self._sendRequest(chunk)
				else:
					for obj in objs:
						self._sendRequest(obj._dumps(), path=obj._path)
	def _sendRequest(self, data:Any, path:str="/", extraHttpHeaders:Optional[Dict[str, str]]=None) -> None:
		payload:bytes
		if isinstance(self.socket, T_HTTPClientSocket):
			extraHttpHeaders = extraHttpHeaders or {}
			extraHttpHeaders["host"] = self.httpHost or "{}:{}".format(self.target[0], self.target[1])
			if self.requestProtocol in ["JSONRPC-2", "JSONRPC-P"]:
				extraHttpHeaders["Content-Type"] = "application/json;charset=utf-8"
				payload = json.dumps(data).encode('utf8')
			elif self.requestProtocol == "FSP":
				extraHttpHeaders["Content-Type"] = "application/fspacker"
				payload = fsPacker.dumps(data, version=self.FSPACKER_VERSION)
			else:
				raise RuntimeError
			self._sendToSocket(payload, path, extraHttpHeaders)
		elif isinstance(self.socket, T_FSPackerClientSocket):
			if self.requestProtocol in ["JSONRPC-2", "JSONRPC-P"]:
				payload = json.dumps(data).encode('utf8')
			elif self.requestProtocol == "FSP":
				payload = fsPacker.dumps(data, version=self.FSPACKER_VERSION)
			else:
				raise RuntimeError
			self._sendToSocket(payload)
		elif isinstance(self.socket, T_OldFSPackerClientSocket):
			if self.requestProtocol == "OldFSProtocol":
				payload = pickle.dumps(data)
			else:
				raise RuntimeError
			self._sendToSocket(payload)
		elif isinstance(self.socket, T_StringClientSocket):
			if self.requestProtocol in ["JSONRPC-2", "JSONRPC-P"]:
				payload = json.dumps(data).encode('utf8')
			elif self.requestProtocol == "FSP":
				payload = fsPacker.dumps(data, version=self.FSPACKER_VERSION).hex().encode('utf8')
			else:
				raise RuntimeError
			self._sendToSocket(payload)
		else:
			raise RuntimeError
	def _sendToSocket(self, payload:bytes, path:str="/", extraHttpHeaders:Dict[str, str]={}) -> None:
		if isinstance(self.socket, T_BaseClientSocket_send_http):
			self.socket.send(payload, path, extraHttpHeaders)
		elif isinstance(self.socket, T_BaseClientSocket_send_default):
			self.socket.send(payload)
		else:
			raise RuntimeError
	def _createSocket(self) -> None:
		if not hasattr(self, "socket"):
			if self.messageProtocol == "HTTP":
				self.socket = HTTPClientSocket(
					client             = self,
					protocol           = self.socketProtocol,
					target             = self.target,
					bind               = self.bind,
					connectTimeout     = self.connectTimeout,
					transferTimeout    = self.transferTimeout,
					ssl                = self.ssl,
					sslHostname        = self.sslHostname,
					extraHeaders       = self.extraHttpHeaders,
					disableCompression = self.disableCompression,
				)
			elif self.messageProtocol == "STR":
				self.socket = StringClientSocket(
					client          = self,
					protocol        = self.socketProtocol,
					target          = self.target,
					bind            = self.bind,
					connectTimeout  = self.connectTimeout,
					transferTimeout = self.transferTimeout,
					ssl             = False,
					sslHostname     = None,
				)
			elif self.messageProtocol == "FSP":
				self.socket = FSPackerClientSocket(
					client          = self,
					protocol        = self.socketProtocol,
					target          = self.target,
					bind            = self.bind,
					connectTimeout  = self.connectTimeout,
					transferTimeout = self.transferTimeout,
					ssl             = False,
					sslHostname     = None,
				)
			elif self.messageProtocol == "OldFSProtocol":
				self.socket = OldFSProtocolClientSocket(
					client          = self,
					protocol        = self.socketProtocol,
					target          = self.target,
					bind            = self.bind,
					connectTimeout  = self.connectTimeout,
					transferTimeout = self.transferTimeout,
					ssl             = False,
					sslHostname     = None,
				)
			else:
				raise RuntimeError
	def _get(self, id:Any) -> None:
		def wh() -> bool:
			return not self.requests[id].isDone()
		if id not in self.requests:
			raise ResponseError("Unknown request ID: {}".format(id))
		self.socketErrors = 0
		while True:
			try:
				self.signal.check()
				if hasattr(self, "socket") and not self.socket.isAlive():
					del self.socket
				if not hasattr(self, "socket"):
					self._createSocket()
					self._connect()
				self.socket.loop(wh)
				break
			except KillSignal:
				raise
			except BaseRPCError:
				if hasattr(self, "socket"):
					del self.socket
				if self.socketErrors < self.retryCount:
					if self.socketErrors > 0:
						self.log.warn("Error while getting response [{} left]", self.retryCount-self.socketErrors)
						if self.retryDelay:
							self.signal.sleep(self.retryDelay)
					self.socketErrors += 1
					continue
				raise
	def _parseResponse(self, payload:bytes, headers:Optional[T_Headers]=None, charset:str="utf8") -> None:
		if self.requestProtocol in ["JSONRPC-2", "JSONRPC-P"]:
			try:
				data = json.loads(payload.decode(charset))
			except:
				raise MessageError("Invalid payload")
			if not isinstance(data, list):
				data = [data]
			for r in data:
				if not isinstance(r, dict):
					raise MessageError("Invalid payload")
				if "id" not in r:
					raise MessageError("Required data missing: id")
				id = r["id"]
				uid = r.get("uid", None)
				isSuccess = not ("error" in r and r["error"])
				if isSuccess:
					if "result" not in r:
						raise MessageError("Required data missing: result")
					result = r["result"]
				else:
					result = r["error"]
				self._parseResult(id, isSuccess, result, uid)
		elif self.requestProtocol == "FSP":
			if self.messageProtocol == "STR":
				payload = bytes.fromhex(payload.decode(charset))
			try:
				_, data = fsPacker.loads(
					payload,
					maxDictSize=0,
					maxOPSize=0,
					maxIndexSize=0,
					recursiveLimit=512,
				)
			except:
				raise MessageError("Invalid payload")
			if not isinstance(data, list):
				data = [data]
			for chunk in data:
				if isinstance(chunk, tuple) and len(chunk) == 4 and \
				(isinstance(chunk[0], (int, str)) or chunk[0] is None) and \
				isinstance(chunk[1], bool) and \
				(isinstance(chunk[3], str) or chunk[3] is None):
					self._parseResult(*chunk)
				else:
					raise MessageError("Invalid payload")
		elif self.requestProtocol == "OldFSProtocol":
			if self.messageProtocol == "STR":
				payload = bytes.fromhex(payload.decode(charset))
			resp = pickle.loads(payload)
			if isinstance(resp, tuple) and len(resp) == 3 and \
			isinstance(resp[0], (int, str)) and \
			isinstance(resp[1], bool) and \
			isinstance(resp[2], bytes):
				self._parseResult(resp[0], resp[1], pickle.loads(decompress(resp[2])), "")
			else:
				raise MessageError("Invalid payload")
	def _parseResult(self, id:Union[int, str], isSuccess:bool, result:Any, uid:str) -> None:
		if id not in self.requests:
			self.log.warn("Got unexpected result id: {}", id)
			return
		obj = self.requests[id]
		self.log.debug("Received result for ID: {} UID: {}".format(id, uid))
		obj._parseResponse( id, isSuccess, result, uid )
	def clear(self) -> None:
		self.requests.clear()
	def clone(self, **kwargs:Any) -> T_Client:
		opts = self.__getstate__()
		opts.update(kwargs)
		return Client(**opts)
	def close(self) -> None:
		if hasattr(self, "socket"):
			self.socket.close()
			del self.socket
		self.clear()
		self.log.debug("Closed")
	__del__ = close
	def request(self, method:str, args:List[Any]=[], kwargs:Dict[str, Any]={}, id:Optional[Union[str, int]]=None,
	path:str="/") -> T_Request:
		if id is None:
			id = self.id
			self.id += 1
		if type(id) not in [int, str]:
			raise RequestError("Request ID can be only str or int")
		if id in self.requests:
			raise RequestError("Request ID already in use: {}".format(id))
		obj = Request(
			self,
			id,
			method,
			args,
			kwargs,
			path,
			self.convertNumbers,
		)
		self.requests[id] = obj
		self.log.info("Request queued: {} [{}]".format(method, id))
		if hasattr(self, "socket"):
			self._sendRequest( obj._dumps() )
		return obj



# Builtin modules
from __future__ import annotations
from abc import ABCMeta, abstractmethod
from selectors import BaseSelector
from ssl import SSLSession
from weakref import WeakValueDictionary
from typing import List, Dict, Tuple, Union, Callable, Optional, Any, NoReturn, Iterable, Iterator
# Third party modules
from fsLogger import T_Logger
from fsSignal import T_Signal
# Local modules
# Program
T_SocketBindAddress_TCPv4 = Tuple[str, int]
T_SocketBindAddress_TCPv6 = Tuple[str, int, int, int]
T_SocketBindAddress = Union[T_SocketBindAddress_TCPv4, T_SocketBindAddress_TCPv6]

class T_Socket(metaclass=ABCMeta):
	def __init__(self, family:int=..., type:int=..., proto:int=..., fileno:Optional[int]=...) -> None: ...
	@abstractmethod
	def fileno(self) -> int: ...
	@abstractmethod
	def do_handshake(self) -> None: ...
	@abstractmethod
	def close(self) -> None: ...
	@abstractmethod
	def setblocking(self, __flag:bool) -> None: ...
	@abstractmethod
	def setsockopt(self, __level:int, __optname:int, __value:Union[int, bytes]) -> None: ...
	@abstractmethod
	def send(self, __data:bytes, __flags:int=...) -> int: ...
	@abstractmethod
	def recv(self, __bufsize:int, __flags:int=...) -> bytes: ...
	@abstractmethod
	def connect_ex(self, __address:Union[Tuple[Any, ...], str]) -> int: ...
	@abstractmethod
	def shutdown(self, __how:int) -> None: ...
	@abstractmethod
	def bind(self, __address:T_SocketBindAddress) -> None: ...
	@abstractmethod
	def getsockname(self) -> T_SocketBindAddress: ...

class SSLContext(metaclass=ABCMeta):
	def wrap_socket(self, sock:T_Socket, server_side:bool=..., do_handshake_on_connect:bool=..., suppress_ragged_eofs:bool=...,
	server_hostname:Optional[str]=..., session:Optional[SSLSession]=...) -> T_Socket: ...

class T_Client(metaclass=ABCMeta):
	max_bulk_request:int
	FSPACKER_VERSION:int
	target:Union[str, Tuple[str, int], Tuple[str, int, int, int]]
	protocol:str
	connTimeout:float
	transferTimeout:float
	retryDelay:float
	retryCount:int
	socketErrors:int
	ssl:bool
	sslHostname:Optional[str]
	httpHost:Optional[str]
	extraHttpHeaders:Dict[str, str]
	disableCompression:bool
	useBulkRequest:bool
	convertNumbers:Optional[str]
	log:T_Logger
	signal:T_Signal
	id:int
	requests:WeakValueDictionary[Optional[Union[str, int]], T_Request]
	socket:Union[T_BaseClientSocket_send_default, T_BaseClientSocket_send_http]
	socketProtocol:str
	messageProtocol:str
	requestProtocol:str
	@abstractmethod
	def _initializeProtocol(self) -> None: ...
	@abstractmethod
	def __enter__(self) -> T_Client: ...
	@abstractmethod
	def __exit__(self, type:Any, value:Any, traceback:Any) -> None: ...
	@abstractmethod
	def __del__(self) -> None: ...
	@abstractmethod
	def __getstate__(self) -> Dict[str, Any]: ...
	@abstractmethod
	def __setstate__(self, states:Dict[str, Any]) -> None: ...
	@abstractmethod
	def _connect(self, sendOlderRequests:bool=...) -> None: ...
	@abstractmethod
	def _sendRequest(self, data:Any, path:str=..., extraHttpHeaders:Optional[Dict[str, str]]=...) -> None: ...
	@abstractmethod
	def _sendToSocket(self, payload:bytes, path:str=..., extraHttpHeaders:Dict[str, str]=...) -> None: ...
	@abstractmethod
	def _createSocket(self) -> None: ...
	@abstractmethod
	def _get(self, id:Any) -> None: ...
	@abstractmethod
	def _parseResponse(self, payload:bytes, headers:Optional[T_Headers]=..., charset:str=...) -> None: ...
	@abstractmethod
	def _parseResult(self, id:Union[int, str], isSuccess:bool, result:Any, uid:str) -> None: ...
	@abstractmethod
	def clear(self) -> None: ...
	@abstractmethod
	def clone(self, **kwargs:Any) -> T_Client: ...
	@abstractmethod
	def close(self) -> None: ...
	@abstractmethod
	def request(self, method:str, args:List[Any]=..., kwargs:Dict[str, Any]=..., id:Optional[Union[str, int]]=...,
	path:str=...) -> T_Request: ...

class T_BaseClientSocket(metaclass=ABCMeta):
	client:T_Client
	protocol:str
	target:Union[str, Tuple[str, int], Tuple[str, int, int, int]]
	connTimeout:float
	transferTimeout:float
	ssl:bool
	sslHostname:Optional[str]
	bind:Optional[T_SocketBindAddress]
	log:T_Logger
	signal:T_Signal
	poll:BaseSelector
	readBuffer:bytes
	writeBuffer:bytes
	connectionStatus:int
	timeoutTimer:float
	sock:T_Socket
	sockFD:Optional[int]
	mask:int
	sslTimer:float
	@abstractmethod
	def _connect(self, initial:bool=False) -> bool: ...
	@abstractmethod
	def _createSocket(self) -> None: ...
	@abstractmethod
	def _doSSLHandshake(self) -> bool: ...
	@abstractmethod
	def _haveRead(self) -> bool: ...
	@abstractmethod
	def _haveWrite(self) -> bool: ...
	@abstractmethod
	def _raiseSocketError(self, err:str) -> NoReturn: ...
	@abstractmethod
	def _raiseMessageError(self, err:str) -> NoReturn: ...
	@abstractmethod
	def _reset(self) -> None: ...
	@abstractmethod
	def _setMask(self, newMask:int) -> None: ...
	@abstractmethod
	def _write(self, data:bytes) -> None: ...
	@abstractmethod
	def close(self) -> None: ...
	@abstractmethod
	def connect(self) -> None: ...
	@abstractmethod
	def _isConnected(self) -> bool: ...
	@abstractmethod
	def isAlive(self) -> bool: ...
	@abstractmethod
	def loop(self, whileFn:Callable[[], bool]) -> None: ...
	@abstractmethod
	def parseReadBuffer(self) -> bool: ...

class T_BaseClientSocket_send_default(T_BaseClientSocket, metaclass=ABCMeta):
	@abstractmethod
	def send(self, payload:bytes) -> None: ...

class T_BaseClientSocket_send_http(T_BaseClientSocket, metaclass=ABCMeta):
	@abstractmethod
	def send(self, payload:bytes=b"", path:str=..., headers:Dict[str, str]=...) -> None: ...

class T_HTTPClientSocket(T_BaseClientSocket_send_http, metaclass=ABCMeta):
	defaultHeaders:Dict[str, str]
	headers:T_Headers
	@abstractmethod
	def send(self, payload:bytes=b"", path:str=..., headers:Dict[str, str]=...) -> None: ...

class T_StringClientSocket(T_BaseClientSocket_send_default, metaclass=ABCMeta):
	pass

class T_FSPackerClientSocket(T_BaseClientSocket_send_default, metaclass=ABCMeta):
	pass

class T_OldFSPackerClientSocket(T_BaseClientSocket_send_default, metaclass=ABCMeta):
	@abstractmethod
	def _readResponse(self) -> Tuple[int, bytes]: ...
	@abstractmethod
	def _encodeRequest(self, buffer:bytes) -> bytes: ...

class T_Headers(metaclass=ABCMeta):
	data:Dict[str, str]
	@abstractmethod
	def __setitem__(self, k:str, v:str) -> None: ...
	@abstractmethod
	def __getitem__(self, k:str) -> str: ...
	@abstractmethod
	def __contains__(self, k:str) -> bool: ...
	@abstractmethod
	def __delitem__(self, k:str) -> None: ...
	@abstractmethod
	def __iter__(self) -> Iterator[str]: ...
	@abstractmethod
	def has_key(self, k:str) -> bool: ...
	@abstractmethod
	def keys(self) -> Iterable[str]: ...
	@abstractmethod
	def values(self) -> Iterable[str]: ...
	@abstractmethod
	def items(self) -> Iterable[Tuple[str, str]]: ...
	@abstractmethod
	def get(self, k:str, default:Optional[str]=None) -> str: ...
	@abstractmethod
	def update(self, m:Dict[str, str]) -> T_Headers: ...
	@abstractmethod
	def dumps(self, extend:Dict[str, str]={}) -> str: ...

class T_Request(metaclass=ABCMeta):
	_client:T_Client
	_id:Optional[Union[str, int]]
	_method:str
	_args:List[Any]
	_kwargs:Dict[Any, Any]
	_path:str
	_convertNumbers:Optional[str]
	_requestTime:float
	_responseTime:float
	_uid:str
	_done:bool
	_success:bool
	_response:Any
	@abstractmethod
	def _get(self) -> None: ...
	@abstractmethod
	def _parseResponse(self, id:Union[int, str], isSuccess:bool, result:Any, uid:str) -> None: ...
	@abstractmethod
	def _dumps(self) -> Any: ...
	@abstractmethod
	def get(self) -> Any: ...
	@abstractmethod
	def getDelay(self) -> float: ...
	@abstractmethod
	def getID(self) -> Any: ...
	@abstractmethod
	def isDone(self) -> bool: ...
	@abstractmethod
	def getUID(self) -> str: ...
	@abstractmethod
	def isSuccess(self) -> bool: ...



# Builtin modules
import os, unittest
from time import monotonic
# Third party modules
from fsLogger import SimpleLogger, Logger
from fsSignal import Signal, SoftSignal
# Local modules
from .. import Client
# Program
class ClientTest(unittest.TestCase):
	signal:Signal
	@classmethod
	def setUpClass(cls) -> None:
		if os.environ.get("DEBUG") == "1":
			SimpleLogger("TRACE")
		cls.signal = Signal()
		return None
	def test_request(self) -> None:
		rootClient = Client(
			"TCPv4:HTTP:JSONRPC-2",
			("api.fusionexplorer.io", 443),
			ssl=True,
			disableCompression=True,
			log=Logger("Client"),
			signal=SoftSignal()
		)
		r = rootClient.request("ping")
		self.assertEqual(r.isDone(), False)
		self.assertEqual(r.isSuccess(), True)
		self.assertEqual(r.isDone(), True)
		self.assertGreater(r.getDelay(), 0)
		self.assertIsInstance(r.get(), str)
		with rootClient as c:
			r = c.request("ping")
			self.assertEqual(r.isDone(), False)
			self.assertEqual(r.isSuccess(), True)
			self.assertEqual(r.isDone(), True)
			self.assertIsInstance(r.get(), str)
			r = c.request("ping", id=88)
			self.assertEqual(r.getID(), 88)
			r = c.request("ping", [0])
			self.assertEqual(r.isSuccess(), False)
			self.assertEqual(r.isDone(), True)
			self.assertEqual(r.get(), {
				"code": -32602,
				"data": "ping() takes 0 positional argument but 1 were given",
				"message": "Invalid params"
			})
			r = c.request("surenoteexists")
			self.assertEqual(r.isSuccess(), False)
			self.assertEqual(r.isDone(), True)
			self.assertEqual(r.get(), {"code": -32601, "data": "surenoteexists", "message": "Method not found"})
			r0 = c.request("ping", id="asd")
			r1 = c.request("ping", id="dsa")
			self.assertIsInstance(r0.get(), str)
			self.assertIsInstance(r1.get(), str)
		with rootClient as c:
			r0 = c.request("ping", id="asd")
			r1 = c.request("ping", id="dsa")
			self.assertIsInstance(r0.get(), str)
			self.assertIsInstance(r1.get(), str)
	def test_convertNumbers_nothing(self) -> None:
		rootClient = Client(
			"TCPv4:HTTP:JSONRPC-2",
			("api.fusionexplorer.io", 443),
			ssl=True,
			disableCompression=True,
			log=Logger("Client"),
			signal=SoftSignal(),
		)
		r = rootClient.request("ping")
		self.assertEqual(r.isSuccess(), True)
		self.assertIsInstance(r.get(), str)
		self.assertRegex(r.get(), r"^[0-9]*$")
		#
		rootClient = Client(
			"TCPv4:HTTP:JSONRPC-P",
			("api.fusionexplorer.io", 443),
			ssl=True,
			disableCompression=True,
			log=Logger("Client"),
			signal=SoftSignal(),
		)
		r = rootClient.request("ping")
		self.assertEqual(r.isSuccess(), True)
		self.assertIsInstance(r.get(), int)
	def test_convertNumbers_default(self) -> None:
		rootClient = Client(
			"TCPv4:HTTP:JSONRPC-2",
			("api.fusionexplorer.io", 443),
			ssl=True,
			disableCompression=True,
			log=Logger("Client"),
			signal=SoftSignal(),
			convertNumbers="default",
		)
		r = rootClient.request("ping")
		self.assertEqual(r.isSuccess(), True)
		self.assertIsInstance(r.get(), str)
		self.assertRegex(r.get(), r"^[0-9]*$")
		#
		rootClient = Client(
			"TCPv4:HTTP:JSONRPC-P",
			("api.fusionexplorer.io", 443),
			ssl=True,
			disableCompression=True,
			log=Logger("Client"),
			signal=SoftSignal(),
			convertNumbers="default",
		)
		r = rootClient.request("ping")
		self.assertEqual(r.isSuccess(), True)
		self.assertIsInstance(r.get(), int)
	def test_convertNumbers_none(self) -> None:
		rootClient = Client(
			"TCPv4:HTTP:JSONRPC-2",
			("api.fusionexplorer.io", 443),
			ssl=True,
			disableCompression=True,
			log=Logger("Client"),
			signal=SoftSignal(),
			convertNumbers="none",
		)
		r = rootClient.request("ping")
		self.assertEqual(r.isSuccess(), True)
		self.assertIsInstance(r.get(), int)
	def test_convertNumbers_str(self) -> None:
		rootClient = Client(
			"TCPv4:HTTP:JSONRPC-P",
			("api.fusionexplorer.io", 443),
			ssl=True,
			disableCompression=True,
			log=Logger("Client"),
			signal=SoftSignal(),
			convertNumbers="str",
		)
		r = rootClient.request("ping")
		self.assertEqual(r.isSuccess(), True)
		self.assertIsInstance(r.get(), str)
		self.assertRegex(r.get(), r"^[0-9]*$")
	def test_convertNumbers_hex(self) -> None:
		rootClient = Client(
			"TCPv4:HTTP:JSONRPC-P",
			("api.fusionexplorer.io", 443),
			ssl=True,
			disableCompression=True,
			log=Logger("Client"),
			signal=SoftSignal(),
			convertNumbers="hex",
		)
		r = rootClient.request("ping")
		self.assertEqual(r.isSuccess(), True)
		self.assertIsInstance(r.get(), str)
		self.assertRegex(r.get(), r"^0x[0-9a-fA-F]*$")
	def test_retry(self) -> None:
		rootClient = Client(
			"TCPv4:HTTP:JSONRPC-P",
			("api.fusionexplorer.io", 443),
			ssl=False,
			disableCompression=True,
			retryCount=2,
			retryDelay=1,
			log=Logger("Client"),
			signal=SoftSignal()
		)
		s = monotonic()
		r = rootClient.request("ping")
		try:
			r.get()
		except:
			pass
		self.assertGreater(monotonic()-s, 1.0)



#!/usr/bin/env python3
import os
try:
	from setuptools import setup # type: ignore
except ImportError:
	from distutils.core import setup

pwd = os.path.abspath(os.path.dirname(__file__))

setup(
	name                          = "python-fsrpcclient",
	version                       = "0.2.0",
	description                   = "Fusion Solutions RPC client",
	keywords                      = "json rpc client fusion solutions fusionsolutions",
	author                        = "Andor `iFA` Rajci - FUSION SOLUTIONS KFT",
	author_email                  = "ifa@fusionsolutions.io",
	url                           = "https://github.com/FusionSolutions/python-fsrpcclient",
	license                       = "GPL-3",
	packages                      = ["fsRPCClient"],
	long_description              = open(os.path.join(pwd, "README.md")).read(),
	long_description_content_type = "text/markdown",
	zip_safe                      = False,
	python_requires               = ">=3.8.0",
	install_requires              = ["python-fslogger~=0.2.0", "python-fssignal~=0.1.0", "python-fspacker~=0.2.0", "lz4"],
	test_suite                    = "fsRPCClient.test",
	package_data                  = { "":["py.typed"] },
	classifiers                   = [ # https://pypi.org/pypi?%3Aaction=list_classifiers
		"Development Status :: 4 - Beta",
		"Topic :: Utilities",
		"Programming Language :: Python :: 3 :: Only",
		"Programming Language :: Python :: 3.8",
		"Programming Language :: Python :: 3.9",
		"License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
	],
)



# Builtin modules
from __future__ import annotations
import zlib
from typing import Dict, Any, List, Tuple, Iterator, Iterable, Optional, Generator, cast
# Local modules
from .abcs import T_Headers
# Program
class Headers(T_Headers):
	def __init__(self, initial:Dict[str, str]={}):
		self.data = {}
		for k, v in initial.items():
			self[k.lower()] = v
	def __setitem__(self, k:str, v:str) -> None:
		self.data[k.lower()] = v
		return None
	def __getitem__(self, k:str) -> str:
		return self.data[k.lower()]
	def __delitem__(self, k:str) -> None:
		del self.data[k.lower()]
	def __iter__(self) -> Iterator[str]:
		return self.data.__iter__()
	def has_key(self, k:str) -> bool:
		return k.lower() in self.data
	__contains__ = has_key
	def keys(self) -> Iterable[str]:
		return self.data.keys()
	def values(self) -> Iterable[str]:
		return self.data.values()
	def items(self) -> Iterable[Tuple[str, str]]:
		return self.data.items()
	def get(self, k:str, default:Optional[str]=None) -> str:
		k = k.lower()
		if k not in self.data:
			return cast(str, default)
		return self.data[k]
	def update(self, m:Dict[str, str]) -> Headers:
		for k, v in m.items():
			self.data[k.lower()] = v
		return self
	def dumps(self, extend:Dict[str, str]={}) -> str:
		h:List[str] = []
		for k, v in self.data.items():
			h.append("{}: {}".format(k, v))
		for k, v in extend.items():
			h.append("{}: {}".format(k.lower(), v))
		return "\r\n".join(h)

class deflate:
	@staticmethod
	def compress(data:bytes) -> bytes:
		o = zlib.compressobj(9, zlib.DEFLATED, -zlib.MAX_WBITS, zlib.DEF_MEM_LEVEL, 0)
		r = o.compress(data)
		r += o.flush()
		return r
	@staticmethod
	def decompress(data:bytes) -> bytes:
		o = zlib.decompressobj(-zlib.MAX_WBITS)
		r = o.decompress(data)
		r += o.flush()
		return r

def hexToBytes(data:str) -> bytes:
	if data[:2].lower() == "0x":
		data = data[2:]
	return bytes.fromhex(data)

def iterSplit(l:List[Any], n:int) -> Generator[Any, List[Any], None]:
	for i in range(0, len(l), n):
		yield(l[i:i+n])
